/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/


#include "nest_collection_manager.h"
#include <string>


// **********************************
// Single NeST collection object
// **********************************
nestCollection::nestCollection( const string &name,
								const string &path,
								ClassAd			*ad )
{
	Clear( );
	Init( name, path, ad );
}

// Create a collection from a ClassAd
nestCollection::nestCollection( ClassAd *ad )
{
	buildCollectionFromAd( ad );
}

// Empty constructor
nestCollection::nestCollection( void )
{
	Clear( );
}

// Destructor
nestCollection::~nestCollection( )
{
}

// Initialize a Collection
NestReplyStatus nestCollection::Init( 
	const string &name,
	const string &path,
	ClassAd	*ad )
{
	if ( ( name.length() == 0 ) || ( path.length() == 0 ) ) {
		return NEST_BAD_PARAMETERS;
	}
	this->name = name;
	this->path = path;
	this->ad = ad;
	if ( ad ) {
		string key = COLLECTION_KEY;
		ad->InsertAttr( "Name", name );
		ad->InsertAttr( "Path", path );
		ad->InsertAttr( "Type", key );
	}
	return NEST_SUCCESS;
}

void
nestCollection::Clear( void )
{
	this->name = "";
	this->path = "";
	this->mtime = 0;
	this->version = "";
	this->comment = "";
	this->ad = NULL;
	this->valid = false;
}

// Build a collection from a ClassAd
bool
nestCollection::buildCollectionFromAd( 
	ClassAd			*ad )
{

    string	key;
	string	str;
	string	usename;

	// Reset the whole thing
	Clear( );

	// Extract the name from the ClassAd if none given
	key = "Name";
	if (!  ad->EvaluateAttrString( key, usename ) ) {
		return false;
	}

	// Extract the path
	key = "Path";
	if (!  ad->EvaluateAttrString( key, str ) ) {
		return false;
	}

	// Ok, we have a ligit collection; build it
	this->name = usename;
	this->path = str;
	this->ad = ad;
	this->valid = true;

	// Finally, extract the other fields & store 'em
	key = "Date";
	Value		val;
	if ( ad->EvaluateAttr( key, val ) ) {
		int	timeval;
		if ( val.IsIntegerValue( timeval ) ) {
		  setMtime( timeval );
		}
	}

	key = "Version";
	if ( ad->EvaluateAttrString( key, str ) ) {
		setVersion( str );
	}

	key = "Comment";
	if ( ad->EvaluateAttrString( key, str ) ) {
		setComment( str );
	}

	return true;
}

NestReplyStatus
nestCollection::setMtime( time_t mtime )
{
	this->mtime = mtime;
	if ( ad  && ( !ad->InsertAttr( "Date", (int) mtime ))  ) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
nestCollection::setVersion( const string &version )
{
	this->version = version;
	if ( ad && ( ! ad->InsertAttr( "Version", version ))  ) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
nestCollection::setVersion( const char *cversion )
{
	string	str;
	if ( cversion ) {
		str = (char *) cversion;
	}
	return setVersion( str );
}

NestReplyStatus
nestCollection::setComment( const string &comment )
{
	this->comment = comment;
	if ( ad && ( ! ad->InsertAttr( "comment", comment ))  ) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
nestCollection::setComment( const char *ccomment )
{
	string	str;
	if ( ccomment ) {
		str = (char *)ccomment;
	}
	return setComment( str );
}


// **********************************
// Collection Manager
// **********************************
nestCollectionManager::nestCollectionManager( void )
{
}

nestCollectionManager::~nestCollectionManager( void )
{
}

int
nestCollectionManager::init( )
{
	// not currently reading anything out of args
    return 0;
}

NestReplyStatus
nestCollectionManager::initialize( ClassAdCollection *resources )
{
	return setupView( resources );
}

NestReplyStatus
nestCollectionManager::recover( ClassAdCollection *resources )
{
	return setupView( resources );
}

// Setup the collection veiw
NestReplyStatus
nestCollectionManager::setupView( ClassAdCollection *resources )
{
	// Store ClassAdCollection information
	this->collection_list = resources;
	this->query.Bind( resources );
	this->collection_view = NEST_COLLECTION_VIEW;

	// Check to see if the "Collection" view already exists
	if ( ! this->collection_list->ViewExists( NEST_COLLECTION_VIEW ) ) {
		ViewName	parent = "root";
		string		view_rank("");
		string		view_partition_expr("");
		string 		view_constraint("");

		// Create the veiw here
		view_constraint = view_constraint + "other.Type == \"" + COLLECTION_KEY + "\"";
		if (!this->collection_list->CreateSubView( this->collection_view,
												   parent,
												   view_constraint,
												   view_rank,
												   view_partition_expr ) ) {
			nest_debug( NEST_DEBUG_ALWAYS,
						"nestCollectionManager Couldn't create view for itself!\n");
			return NEST_CLASSAD_ERROR;
		}
	}
	return NEST_SUCCESS;
}

NestReplyStatus
nestCollectionManager::createAssociation( const string &name,
                                          const string &path )
{
	// If it already exists, return an error
    if ( checkName( name ) ) {
        return NEST_ALREADY_EXISTS;
    }

	// Build a collection
	ClassAd			*ad = new ClassAd;
	nestCollection	*col = new nestCollection( name, path, ad );
	delete col;

	string	key = COLLECTION_KEY + name;
    if (!this->collection_list->AddClassAd( key, ad )) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "AddClassAd failed - key = %s - CondorErrMsg = %s \n",
                    key.c_str(),
                    CondorErrMsg.c_str() );

        return NEST_CLASSAD_ERROR;
    }

	
	
    return NEST_SUCCESS;
}

NestReplyStatus
nestCollectionManager::removeAssociation( const string &name )
{
    if ( ! checkName( name ) ) {
        return NEST_LOCAL_FILE_NOT_FOUND;
    }

    string key = COLLECTION_KEY + name;
    if (!this->collection_list->RemoveClassAd( key )) {
        return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

bool
nestCollectionManager::checkName( const string &name )
{
    string key = COLLECTION_KEY + name;

    ClassAd *ad = this->collection_list->GetClassAd( key );
    if (! ad) {
        return false;
    }

	return true;
}

nestCollection *
nestCollectionManager::getCollection( const string &name )
{
    string key = COLLECTION_KEY + name;

    ClassAd *ad = this->collection_list->GetClassAd(key);
    if (!ad) {
        return NULL;
    }

	// And, build a collection from it
    return new nestCollection( ad );
}

const string *
nestCollectionManager::getPath( const string &name )
{
    string key = COLLECTION_KEY;
    key = key + name;

    ClassAd *ad = this->collection_list->GetClassAd(key);
    if (!ad) {
        return NULL;
    }

	key = "Path";
	string	*str = new string;
	if (! ad->EvaluateAttrString( key, *str ) ) {
		return NULL;
	}

	return str;
}

NestReplyStatus
nestCollectionManager::updateMtime( nestCollection *col, time_t mtime )
{
	NestReplyStatus	status = col->setMtime( mtime );
	if ( status != NEST_SUCCESS ) {
		return status;
	}
	return update( col );
}

NestReplyStatus
nestCollectionManager::updateVersion( nestCollection *col, const string &version )
{
	NestReplyStatus	status = col->setVersion( version );
	if ( status != NEST_SUCCESS ) {
		return status;
	}
	return update( col );
}

NestReplyStatus
nestCollectionManager::updateComment( nestCollection *col, const string &comment )
{
	NestReplyStatus	status = col->setVersion( comment );
	if ( status != NEST_SUCCESS ) {
		return status;
	}
	return update( col );
}

NestReplyStatus
nestCollectionManager::update( nestCollection *col )
{
	ClassAd	*ad = col->getAd( );
	string tmp( (char *)col->getName( ) );
    string key = COLLECTION_KEY + tmp;
    if (! this->collection_list->UpdateClassAd( key, ad ) ) {
		return NEST_CLASSAD_ERROR;
	}
	return NEST_SUCCESS;
}

NestReplyStatus
nestCollectionManager::VisitAllCollections( nestCollectionVisitor &visitor )
{

	// Setup the query
    string constraint = "";
    ExprTree    *constraint_tree = this->parser.ParseExpression( constraint );

	// Setup the query
    if ( ! this->query.Query( this->collection_view, constraint_tree ) ) {
		nest_debug( NEST_DEBUG_ALWAYS, "Collection query failed: %s\n",
					CondorErrMsg.c_str() );
        return NEST_CLASSAD_ERROR;
	}
    this->query.ToFirst();
    string query_key;

	// And, start the query
    if (! this->query.Current( query_key ) ) {
		return NEST_SUCCESS;
    }

	// Loop through all of the matching ads
    do {
		ClassAd	*ad = this->collection_list->GetClassAd( query_key );
        if (! ad) {
			nest_debug( NEST_DEBUG_ALWAYS, "No ad\n" );
            continue;
        }
		nestCollection	*collection = new nestCollection( ad );
        if (! collection) {
			nest_debug( NEST_DEBUG_ALWAYS,
						"Failed building collection from Ad\n" );
            continue;
        }

		visitor.visit( *collection );
    } while ( this->query.Next( query_key ) );

	return NEST_SUCCESS;
}
