/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "nest_debug.h"
#include "nest_protocol.h"

bool daemonize = false;
nestProtocol *logfile = NULL;


int nest_debug_level = NEST_DEBUG_MIN;
FILE * nest_debug_output = stderr;

#define DEBUG_DELIMITER ","

struct debug_s
{
    char *name;
    int  value;
};

struct debug_s debug_list[] = {
    { "concurrency",    NEST_DEBUG_CONCURRENCY },

    { "user",           NEST_DEBUG_USER_MNGMNT },
    { "group",          NEST_DEBUG_GROUP_MNGMNT },
    { "account",        NEST_DEBUG_ACCT_MNGMNT },

    { "lots",           NEST_DEBUG_LOTS },

    { "acl",            NEST_DEBUG_ACL },
    { "debug",          NEST_DEBUG_DEBUG },
    { "transfer",       NEST_DEBUG_TRANSFER },

    { "client",         NEST_DEBUG_CLIENT },
    { "storage",        NEST_DEBUG_STORAGE },
    { "server",         NEST_DEBUG_SERVER },
    { "admin",          NEST_DEBUG_ADMIN },

    { "auth",           NEST_DEBUG_AUTH },
    { "fd",             NEST_DEBUG_FD },
    { "proxy",          NEST_DEBUG_PROXY },

    { "flock",          NEST_DEBUG_FLOCK },
    { "malloc",         NEST_DEBUG_MALLOC },

    { "chirp",          NEST_DEBUG_CHIRP },
    { "http",           NEST_DEBUG_HTTP },
    { "protpref",       NEST_DEBUG_PROTOCOL_SCHED },
    { "protocol",       NEST_DEBUG_PROTOCOL },

    { "always",         NEST_DEBUG_ALWAYS },
    { "never",          NEST_DEBUG_NEVER }, 
    { 0, 0 }
};

NestReplyStatus
nest_debug_init( nestProtocol *file, bool daemon )
{
    logfile = file;
    daemonize = daemon;
    return NEST_SUCCESS;
}

long
get_nest_debug_level( )
{
    return nest_debug_level;
}

void
set_nest_debug_level( long level )
{
    nest_debug_level = level;
}

void
add_nest_debug_level( long new_level )
{
    if ( new_level == NEST_DEBUG_NEVER ) {
        nest_debug_level = NEST_DEBUG_NEVER;
    } else {
        nest_debug_level |= new_level;
    }
}

NestReplyStatus
add_nest_debug_level( const char *debug_values )
{
    char *debug_values_copy = NestStrdup(debug_values);
    if (debug_values_copy != NULL) {
        char *debug = strtok( debug_values_copy, DEBUG_DELIMITER );
        struct debug_s * debugs;
        
        while( debug ) {
            for( debugs = debug_list; debugs->name != NULL; debugs++ ) {
                if ( ! strcmp( debug, debugs->name ) ) {
                    nest_debug( NEST_DEBUG_DEBUG, "Setting debug for %s\n",
                                debugs->name );
                    add_nest_debug_level( debugs->value );
                    break;
                }
            }
            if ( debugs->name == NULL ) {
                return NEST_UNKNOWN_DEBUG_VALUE;
            }
            debug = strtok( NULL, DEBUG_DELIMITER );
        }
        NestFree(debug_values_copy);
    }

    return NEST_SUCCESS;
}

int
nest_print( const char *initial_string, const char *format, va_list args )
{
	static time_t	start_time = 0;
    char output[NEST_MESSAGESIZE];
	int	 ilen;

	// Get the time, set start_time if first time...
    time_t ticks = time( NULL );
	if ( 0 == start_time ) {
		start_time = ticks;
	}

	// Build the time string
	struct tm *tb = localtime( &ticks );
	char timestr[32];
	snprintf( timestr, sizeof( timestr),
			  "%d/%d %02d:%02d:%02d =%lds",
			  tb->tm_mon, tb->tm_mday,
			  tb->tm_hour, tb->tm_min, tb->tm_sec,
			  ticks - start_time);

    if ( initial_string ) {
        ilen = NestSnprintf( output, NEST_MESSAGESIZE, "%s:%s: ",
							 timestr, initial_string );
    } else {
        ilen = NestSnprintf( output, NEST_MESSAGESIZE, "%s: ",
							 timestr );
    }
	if ( ilen < 0 ) {
		ilen = 0;
	}

        // TODO: count the number of args in the format string
        // and make sure it is the same as the number of args in the va_list.
    
    int retval = vsnprintf( output + ilen,
                            NEST_MESSAGESIZE - ilen, 
                            format, 
                            args );
        // TODO: how do I check the return value of stupid vsnprintf

    if ( daemonize ) {
        syslog( LOG_NOTICE | LOG_USER, "%s", output );
    } else if ( logfile != NULL ) {
        logfile->Sendn( output, NestStrlen( output ) );
    } else {
        fprintf( nest_debug_output, "%s", output );
        fflush(  nest_debug_output );
    }
    return retval;
}

int 
nest_printf( const char *format, ... )
{
    va_list args;
    va_start( args, format );
    int ret = nest_print( NULL, format, args );
    va_end( args );
    return ret;
}


#ifdef DEBUG

bool 
check_nest_debug_status( long level )
{
    return (nest_debug_level & level);
}

int
nest_debug( long level, const char *format, ... )
{
    int ret = 0;
    if ( check_nest_debug_status( level ) ) {
        va_list args;
        va_start( args, format );
        ret = nest_print( NULL, format, args );
        va_end( args );
    }
    return ret;
}

#else

int 
nest_debug( long level, const char *format, ... )
{
    return 1;
}

#endif

int
nest_warn( const char *filename,
		   const char *func_name,
		   const char *format, ... )
{
    char output[NEST_MESSAGESIZE];
    int ret = 0;
    if ( ! getenv( "NEST_SILENCE" ) ) {
        va_list args;
        va_start( args, format );
        NestSnprintf(output, NEST_MESSAGESIZE,
					 "WARNING @ '%s:%s': ", filename, func_name );
        ret = nest_print( output, format, args );
        va_end( args );
    }
    return ret;
}

void
nest_die( const char *format, ... )
{
    va_list args;
    va_start( args, format );
    nest_print( "NeST Fatal Error: ", format, args );
    va_end( args );
    _exit( 1 );
}


void
nest_perror( const char *msg )
{
    nest_printf( "%s: %s\n", msg, strerror( errno ) );
}
