/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include "nest_dsi_config.h"
#include "nest_dsi_logger.h"
#include "nest_shared.h"

// If we don't have recursive mutex's, don't do thread locking at all..
#ifndef PTHREAD_MUTEX_RECURSIVE
# undef USE_PTHREADS
#endif

typedef struct {
	const char	*label;
	unsigned	flags;
} LogFlags;

static LogFlags logFlags [] =
{
	{ "fulldebug",		NEST_DSI_LOG_FULLDEBUG },
	{ "session",		NEST_DSI_LOG_SESSION },
	{ "transfer",		NEST_DSI_LOG_XFER },
	{ "xfer",			NEST_DSI_LOG_XFER },
	{ "transfer_full",	NEST_DSI_LOG_XFER_FULL },
	{ "xfer_full",		NEST_DSI_LOG_XFER_FULL },
	{ "commands",		NEST_DSI_LOG_COMMANDS },
	{ "stat",			NEST_DSI_LOG_STAT },
	{ "finish",			NEST_DSI_LOG_FINISH },
	{ "all",			NEST_DSI_LOG_ALL },
	{ NULL, 0 },
};

nestDsiLogger::nestDsiLogger( const nestDsiConfig *config )
{
	const char	*file = config->getLogFile( );

	// Default -- use stdout
	if ( NULL == file ) {
		file = "stdout";
	}

	// Now, open up the file...
	if ( ! strcmp( file, "stdout" ) ) {
		log_type = LOG_STDOUT;
	}
	else if ( ! strcmp( file, "stderr" ) ) {
		log_type = LOG_STDERR;
	} else {
		FILE *fp = fopen( file, "a" );
		if ( NULL == fp ) {
			file = "stdout";
			log_type = LOG_STDOUT;
		} else {
			chmod( file, 0666 );
			fclose( fp );
			log_type = LOG_FILE;
		}
	}

	// Store off what we've figured out...
	this->log_file = strdup( file );
	char *tmp = ( char *)malloc( strlen( file ) + 5 );
	strcpy( tmp, file );
	strcat( tmp, ".old" );
	this->old_file = tmp;
	this->rotating = false;

	// Get the max log size
	max_size = config->getLogMaxSize( );
	if ( max_size < 1024 ) {
		max_size = 1024;
	}

	// Now, look at the log flags
	log_flags = NEST_DSI_LOG_ALWAYS;
	const char	*flagstr = config->getLogFlags( );
	if ( flagstr ) {
		char	*flagcopy = strdup( flagstr );
		char	*cur;
		char	*flag = strtok_r( flagcopy, ", ", &cur );
		while( flag ) {
			LogFlags	*flagent;
			for( flagent = logFlags;  flagent->label;  flagent++ )
			{
				if ( ! strcasecmp( flag, flagent->label ) )
				{
					log_flags |= flagent->flags;
				}
			}
			flag = strtok_r( NULL, ", ", &cur );
		}
		free( flagcopy );
	}

# ifdef USE_PTHREADS
	pthread_mutexattr_t   mta;
	pthread_mutexattr_init( &mta );
	pthread_mutexattr_settype(&mta, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init( &mutex, &mta );
# endif
}

nestDsiLogger::~nestDsiLogger( void )
{
	if ( log_file ) {
		free( (void *) log_file );
	}
}

void
nestDsiLogger::logEvent( unsigned flags, const char *format, ... )
{
    va_list args;
    va_start( args, format );

    logEventVa( flags, format, args );

    va_end( args );
}

void
nestDsiLogger::Lock( void )
{
# if USE_PTHREADS
	pthread_mutex_lock( &mutex );
# endif
}

void
nestDsiLogger::Unlock( void )
{
# if USE_PTHREADS
    pthread_mutex_unlock( &mutex );
# endif
}

void
nestDsiLogger::logEventVa( unsigned flags,
						   const char *format,
						   va_list args )
{

	// Should we log this?
	if ( ! ( flags & log_flags ) )
	{
		return;
	}

	// The file to use...
	FILE	*fp;

	// Lock the mutex
	Lock( );

	// Open the log file
	if ( LOG_FILE == log_type ) {
		fp = fopen( log_file, "a" );
		if ( fp ) {
			chmod( log_file, 0666 );
		} else {
			// Can't write to the file -- switch to stdout
			log_type = LOG_STDOUT;
		}
	}

	// STDOUT/ERR
	if ( LOG_STDOUT == log_type ) {
		fp = stdout;
	} else if ( LOG_STDERR == log_type ) {
		fp = stderr;
	}

	// Log possible errors
	char		erbuf[1024];
	erbuf[0] = '\0';

	// And, log it
	time_t		now = time( NULL );
	char		timestr[32];
	ctime_r( &now, timestr );
	int			len = strlen( timestr );
	if ( timestr[len-1] == '\n' ) {
		timestr[len-1] = '\0';
	}
	char	buffer[2048];
	vsnprintf( buffer, sizeof(buffer), format, args );
	fprintf( fp, "%s %s", timestr, buffer );

	// Close the log file
	if ( LOG_FILE == log_type ) {
		fclose( fp );
		if ( ! rotating ) {		 	// This is safe: we own the semaphore...
			struct stat	sb;
			if ( stat( log_file, &sb ) ) {
				fprintf( stderr, "Error stating log file '%s': '%s' (%d)\n",
						 log_file, strerror( errno ), errno );
				goto DONE;
			}
			if ( sb.st_size >= max_size ) {
				rotating = true;	// This is safe: we own the semaphore...

				if ( remove( old_file ) ) {
					if ( ( errno != ENOENT ) && ( sb.st_uid == getuid() ) ) {
						logEvent( NEST_DSI_LOG_ALWAYS,
								  "Error removing old log file '%s'"
								  ": '%s' (%d)\n",
								  old_file, strerror( errno ), errno );
						logEvent( NEST_DSI_LOG_ALWAYS,
								  "My UID/EUID = %d/%d\n",
								  getuid(), geteuid() );
					}
				}
				if ( ! rename( log_file, old_file ) ) {
					char	tmp[32];
					logEvent( NEST_DSI_LOG_ALWAYS,
							  "Log rotated to %s (size limit is %s)\n",
							  old_file,
							  NestFormatSizeString(max_size,
												   tmp, sizeof(tmp) )  );
				} else if ( sb.st_uid == getuid() ) {
					snprintf( erbuf, sizeof( erbuf ),
							  "Can't rotate log file '%s'->'%s': '%s' (%d)\n",
							  log_file, old_file, strerror( errno ), errno );
					fprintf( stderr, erbuf );
					logEvent( NEST_DSI_LOG_ALWAYS, erbuf );
				}
				rotating = false;	// This is safe: we own the semaphore...
			}
		}
	}

	// Unlock the mutex
  DONE:
	Unlock( );
}
