/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_io_manager.h"
#include "nest_filetransfer.h"

nestFiletransfer::nestFiletransfer( nestProtocol *src, nestProtocol *dst,
        long length, char *buffer, long buffersize, int type )
{
    this->type      = type;
    this->initlen   = length; 
    this->src       = src;
    this->dst       = dst;
    this->buffer    = buffer;
    this->buflength = (long) buffersize;
    this->end       = NULL;
    this->started   = false;
    this->allocation = 0;

        // measurements
    NestMemset( &info, 0, sizeof( struct IO_Measurement ) );
}

void
nestFiletransfer::setLimit( long allocation ) {
    this->allocation = allocation;
}

long
nestFiletransfer::getLimit( ) {
    return this->allocation;
}

long
nestFiletransfer::getBytesTransferred( ) {
    assert( started );
    return info.total_bytes; 
}

int
nestFiletransfer::getType() {
    return type;
}

bool
nestFiletransfer::isIncoming() {
    return ( type == THIRD_PARTY_GET || type == CLIENT_SEND_FILE );
}

void
nestFiletransfer::setBytesTransferred( long bytes ) {
    info.total_bytes = bytes;
}

long 
nestFiletransfer::getInitialLength( ) {
    return initlen;
}
    

nestFiletransfer::~nestFiletransfer( ) {
    assert( started );
    if ( end != NULL ) {
        NestFree( end );
    }
}

struct IO_Measurement *
nestFiletransfer::getMeasurements( ) {
    assert( started );
    if (     info.total_bytes     < 0 || info.first_in_time < 0 
         ||  info.first_out_time  < 0 || info.total_in_time < 0
         ||  info.total_out_time  < 0 || info.first_in_size < 0
         ||  info.first_out_size  < 0 ) 
    {
        nest_warn( __FILE__, __FUNCTION__, "Bad measurements in %s: %d\n", __FILE__, __LINE__ );
    }
    return &info;
}

void 
nestFiletransfer::setMeasurements( struct IO_Measurement *measurements ) {
    memcpy( &info, measurements, sizeof( struct IO_Measurement ) );
}

NestReplyStatus 
nestFiletransfer::startTransfer( ) {
    NestGettimeofday( &start );
    started = true;
    return NEST_SUCCESS;
}

NestReplyStatus 
nestFiletransfer::endTransfer( NestReplyStatus status ) {
    assert( started );
    if ( end == NULL ) {
        end = ( struct timeval * ) NestMalloc( sizeof( struct timeval ) );
    }
    NestGettimeofday( end );
    return NEST_SUCCESS;
}

long
nestFiletransfer::getLatency( ) {
    assert( started );
        // these extra timevals allow us to check progress
    struct timeval now, *use;
    if ( end == NULL ) {
        NestGettimeofday( &now );
        use = &now;
    } else {
        use = end;
    }
    return NestTimediff( &start, use, __FUNCTION__ );
}

float
nestFiletransfer::getBandwidth( ) {
    return NestBandwidth( info.total_bytes, getLatency() ); 
}

// transfer initlen bytes from src to dst
// however initlen can be set to ENTIRE_FILE (-1) which means that the transfer
// initlen is unknown
NestReplyStatus
nestFiletransfer::BlockingTransfer( void )
{

    // TODO:  add code for asking the protocol objects for their preferred
    // block sizes

    bool total_known; 
    bool limit_set;
    long remaining;

    if ( initlen != ENTIRE_FILE ) {
        remaining = initlen;
        total_known = true;
        limit_set   = false;
    } else if ( allocation > 0 ) {
        remaining = allocation;
        total_known = false;
        limit_set   = true;
    } else {
        remaining = buflength;
        total_known = false;
        limit_set   = false;
    }

    long bytesToTransfer, bytesRead;
    struct timeval start, end, timer;


    NestGettimeofday( &start );
    nest_debug( NEST_DEBUG_TRANSFER,
				"Blocking transfer: %ld length\n", initlen );

        // TODO: re-add the ability to test bandwidth
    if ( src == NULL || dst == NULL ) return NEST_NOT_YET_IMPLEMENTED;

    int zero_read_attempts = 10;

        // use remaining as a break condition if the initlen is known
        // otherwise just keep transferring until Receive returns 0
    while( remaining > 0 ) {

        /*
        if ( initlen == ENTIRE_FILE ) {
            bytesToTransfer = buflength;
        } else {
            bytesToTransfer = ( remaining < buflength ) 
                ? remaining : buflength;    
        }
        */
        bytesToTransfer = ( remaining < buflength ) ? remaining : buflength;

            // receive
        NestGettimeofday( &timer );
        nest_debug( NEST_DEBUG_TRANSFER, 
                "Blocking Transfer trying to receive %d\n",
                bytesToTransfer );

        if ( total_known ) {
            bytesRead = src->Receiven( buffer, bytesToTransfer );
        } else {
            bytesRead = src->Receive( buffer, bytesToTransfer );
        }
        if ( bytesRead < 0 ) {
            nest_debug( NEST_DEBUG_TRANSFER, 
                    "Blocking transfer recv failed: %s\n", strerror( errno ) );
            return ErrnoToNest( errno );
        }
        NestGettimeofday( &end );

            // collect in info
        long interval = NestTimediff( &timer, &end, __FUNCTION__ );
        if ( interval < 0 ) {
            nest_warn( __FILE__, __FUNCTION__,
					   "Negative interval %ld in %s for reading %ld bytes\n",
					   interval, __FUNCTION__, bytesRead );
        }
        info.total_in_time += interval; 
        if ( info.first_in_time == 0 ) {
            info.first_in_time = interval; 
            info.first_in_size = bytesRead;
        }

            // then send if bytesRead greater than 0
        if ( bytesRead > 0 ) {
            NestGettimeofday( &timer );
			nest_debug( NEST_DEBUG_TRANSFER,
						"BlockingTransfer: Read %d bytes\n", bytesRead );
			nest_debug( NEST_DEBUG_TRANSFER,
						"BlockingTransfer: dst=%p\n", dst );
			nest_debug( NEST_DEBUG_TRANSFER,
						"BlockingTransfer: buf=%p\n", buffer );
            if ( dst->Sendn( buffer, bytesRead ) != bytesRead ) {
                nest_warn( __FILE__, __FUNCTION__, 
						   "Blocking transfer sendn failed: %s\n",
						   strerror( errno ) );
                return ErrnoToNest( errno );
            }
            NestGettimeofday( &end );

            nest_debug( NEST_DEBUG_TRANSFER, "Transferred %ld.\n", bytesRead );

                // collect out info
            interval = NestTimediff( &timer, &end, __FUNCTION__ );
            if ( interval < 0 ) {
                nest_warn( __FILE__, __FUNCTION__,
						   "Negative interval %ld in %s"
						   " for sending %ld bytes\n",
                        interval, __FUNCTION__, bytesRead );
            }
            info.total_out_time += interval; 
            if ( info.first_out_time == 0 ) {
                info.first_out_time = interval; 
                info.first_out_size = bytesRead;
            }
        }

            // then adjust
        info.total_bytes += bytesRead;

            // unknown length transfers are provided an initial allocation
        //if ( initlen != ENTIRE_FILE ) 
        if ( total_known || limit_set ) {
            remaining -= bytesRead;
        }

        if ( bytesRead == 0 ) {
            if ( initlen == ENTIRE_FILE ) {
                    // assume EOF and break early
                nest_debug( NEST_DEBUG_TRANSFER,
							"Received 0 bytes for file of "
							"unknown size.  Assuming EOF.\n" );
                break;
            } else if ( initlen != ENTIRE_FILE ) {
                off_t offset = 0;
                if ( src->hasSocket() ) {
                    offset = NestLseek( src->getSocket(), (off_t)0, SEEK_CUR );    
                }
                nest_warn( __FILE__, __FUNCTION__,
						   "Recvd 0 out of %d bytes in %s from %s:%d "
						   "for translen %ld (%ld so far - offset %ld)."
						   "  Assuming client was lost. %d more attempts.\n",
                    bytesToTransfer,
                    __FUNCTION__, 
                    src->getName(),
                    src->getSocket(), 
                    initlen,
                    info.total_bytes,
                    offset,
                    zero_read_attempts );
                zero_read_attempts--;
                if ( zero_read_attempts <= 0 ) {
                    return NEST_NO_CONNECTION; 
                } else {
                    NestNanosleep( 10000 * (10 - zero_read_attempts) );
                }
            }
        }
    }

        // sanity check
    /*
     * FIX ME: Should NeST die just because a file transferred less than
     * what it advertised?
     */
    if ( initlen != ENTIRE_FILE && remaining != 0 ) {
        nest_warn(
			__FILE__, __FUNCTION__,
			"Badness in %s.%s. Expected len was %ld, but %ld remains.\n", 
			__FILE__, __FUNCTION__, initlen, remaining );
    }

        // check to see if we've hit any restrictions in terms of how much data 
        // we can put into the nest...
    NestReplyStatus status = NEST_SUCCESS;
    if ( initlen == ENTIRE_FILE && allocation > 0 && remaining == 0 ) {
        status = NEST_LOT_NOT_ENOUGH;
        nest_debug( NEST_DEBUG_LOTS, "Blocking transfer of %ld bytes is "
                "successful, but need to request additional lot allocation.\n",
                allocation );
    }


    endTransfer( status );
    NestGettimeofday( &end );
    nest_debug( NEST_DEBUG_TRANSFER, "BlockingTransfer %ld bytes (%.3f)\n", 
            info.total_bytes, getBandwidth( ) ); 
    return status;
}
