/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <assert.h>

#include "nest_speak.h"
#include "nest_debug.h"

// used to know how much total data we transferred
nestFileSize transfer_len = 0;
nestFileSize so_far = 0; 
void
quietProgress( nestFileSize transferred, nestFileSize total, void *args )
{
    so_far = transferred; 
    transfer_len = total;
}

void
verboseProgress( nestFileSize transferred, nestFileSize total, void *args )
{
    NestTransferInfo *transfer = ( NestTransferInfo *) args;
    struct timeval now;
    NestGettimeofday( &now );
    assert( transfer );
    so_far = transferred;
    transfer_len = total;

    long micros = NestTimediff( &transfer->starttime, &now, __FUNCTION__ );
    double bw = NestBandwidth( transferred, micros );
    int percent = (int)( ( transferred * 100.0 ) / total );

        // print the progress info
    fprintf( stdout, "\r%3d%% %10lld / %10lld (%6.2f MB/s) ", 
			 percent, transferred, total, bw );
    fflush(  stdout );
}


bool
Match( char *one, char *two )
{
    return ( !strcasecmp( one, two ) );
}

void
Usage( char *prog, int line )
{
    fprintf(
		stderr, 
		"Usage (%d): %s: [-a] [-v] [-p #] put/get server src_path dst_path\n"
		"\t-a anonymous connection\n"
		"\t-v verbose; shows progress\n"
		"\t-p specify a port on which to connect to server\n"
		"\t-q silent\n",
		line, prog);
    exit( 1 );
}

void
PrintProxy( char *proxy, void *args )
{
    int *found = (int *)args;
    (*found)++;
    printf( "%s\n", proxy );
}

int
ProxyQuery( NestConnection con)
{
    int found = 0;
    NestReplyStatus stat = NestProxyQuery( con, PrintProxy, &found );
    if ( stat != NEST_SUCCESS ) return -1;
    else return 0;
}

int 
main( int argc, char **argv )
{
    char *server, *srcpath, *dstpath, *command;
    int arg = 1; 
    NestProgressFunction *progress = quietProgress;
    NestTransferInfo *transferInfo =  NULL; 
    NestReplyStatus ( *transfer ) ( const char *,
									const char *,
									NestConnection, 
									NestProgressFunction *,
									void * ) = NULL;
    int port = NEST_SERVER_PORT;
    int quiet = 0;
    int anon  = 0;
    int c;

    char *prog = argv[0];

    while( ( c = getopt( argc, argv, "avqp:" )) != EOF ) {
        switch( c ) {
        case 'a':
            anon = 1;
            break;
        case 'v':
            progress = verboseProgress;
            transferInfo = NestmakeTransferInfo();
            break;
        case 'q':
            quiet = 1;
            break;
        case 'p':    
            port = atoi( optarg );
            break;
        default:
            Usage( prog, __LINE__ );
        }
    }

    //if ( optind + 4 != argc ) {
    //    Usage( prog, __LINE__ );
   // }
    arg = optind;

        // next is get or put or proxys, then comes server name
    command = argv[arg++];
    server = argv[arg++];
        // try to connect to the server first
    NestConnection connection;
    NestReplyStatus status;
    if ( anon ) status = NestOpenAnonymously( connection, server, port );
    else status = NestOpenConnection( connection, server, port ); 
    if ( status != NEST_SUCCESS ) {
        fprintf( stderr, "Could not connect to NeST server %s: %s\n",
                server, NestErrorString( status ) );
        exit( 1 );
        exit( -1 );
    }
    
    if (    ! Match( command, "put" ) || 
            ! Match( command, "get" ) ||
            ! Match( command, "proxys" ) ) {
        if ( Match( command, "put" ) ) transfer = NestSendFile;
        else if ( Match( command, "proxys" ) ) return ProxyQuery( connection );
        else                           transfer = NestReceiveFile;
    } else Usage( prog, __LINE__ );
    
        // next is server, srcpath, dstpath
    srcpath = argv[arg++];
    dstpath = argv[arg++];

        // do transfer 
    struct timeval start, end;
    NestGettimeofday( &start );
    status = transfer( srcpath, dstpath, connection, progress, transferInfo );
    NestGettimeofday( &end );

        // report what happened
    if ( progress == verboseProgress ) printf( "\n" );
    if ( so_far != transfer_len ) {
        assert( so_far < transfer_len );
        nest_warn( __FILE__, __FUNCTION__,
				   "Did not receive full transfer. Expected %ld, recvd %ld.\n",
				   transfer_len, so_far );
    }
    if ( status != NEST_SUCCESS ) {
        nest_warn( __FILE__, __FUNCTION__, "Could not %s %s: %s\n",
                command, srcpath, NestErrorString( status ) );
    } else if ( !quiet ) {
        printf( "%lld bytes, %.5f MB/s\n",
                so_far, 
                NestBandwidth(transfer_len, &start, &end, __FUNCTION__));
    }
    
    if ( transferInfo ) NestFree( transferInfo );
    return 0;
}
