/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_proxy.h"

#include "nest_io_manager.h"
#include "nest_collection_manager.h"
#include "nest_storage_manager.h"
#include "nest_client_pool.h"
#include "nest_protocol_manager.h"
#include "nest_advertiser.h"
#include "nest_protocol_advertise.h"
#include "nest_request_handler.h"
#include "nest_config.h"
#include "nest_hash_table.h"
#include "nest_user.h"
#include "nest_version.h"
#include "nest_server.h"
#include "nest_main.h"

#include <sys/utsname.h>
#include <rpc/rpc.h>

    // handlers for the various services provided to clients
static void setNewGroup( void );

    // other misc helper functions
static void init( int argc, const char *argv[] );
static void installSignalHandlers( void );
static void sigChild( int signal, siginfo_t *info, void *context );
static void debug_init( bool Daemonize );
static void daemon_init( const char *pname, int facility );

    // the pid of the main server thread
	// used to make sure that only main server thread performs
	// protected duties
extern int admin_pid;

    // needs to be global now for taskDone
static fd_set writefds, readfds, r_readyfds, w_readyfds;

nestServer	*nest_server;


int
main( int argc, const char *argv[] )
{
	init( argc, argv );

        // enter the main server loop after doing first RegularMaintenance
        // first handle any messages from our dispatched worker processes
        // then check for any new incoming client requests
    nest_server->RegularMaintenance();
    nestClientPool *request_queue = new nestClientPool( );
	nest_printf( "Initializing done.  Ready to service requests.\n" );
    while( true ) {

		int	max_fd = nest_server->getMaxfd() + 1;

            // merge rpc readlist with our own, then enter select loop
        r_readyfds = svc_fdset;
        NestMergeFdsets( &r_readyfds, &readfds, max_fd );
        nest_debug( NEST_DEBUG_FD, "NestRead set ( %x ): ", &readfds );
        NestFd_print( NEST_DEBUG_FD, &readfds, max_fd );

        w_readyfds = writefds;
        nest_debug( NEST_DEBUG_FD,
                "maxfd is %d, entering select\n", max_fd );
        int readies = NestSelect(max_fd, &r_readyfds,&w_readyfds, NULL,
								 nest_server->getNextTimeout( ) );

        if ( readies <= 0 ) {
            if ( readies == 0 ) {
                    // TODO need slow timeouts for regular maintenance
                    // and fast timeouts for scheduling
                nest_server->RegularMaintenance();
            } else nest_warn( __FILE__, __FUNCTION__, 
                    "select failed: %s\n", strerror( errno ) );
            continue;
        }

        nest_debug( NEST_DEBUG_FD, "Woke from select with %d readies.\n",
					readies );

            // then accept any new clients - listen on each protocol socket
        for( int sock = protocol_manager->firstProtocolSocket();
             sock != protocol_manager->lastProtocolSocket();
             sock  = protocol_manager->nextProtocolSocket() )
        {
                // ignore any sockets that were unable to listen correctly
            if ( sock <= 0 ) {
				continue;
			}

                // some protocols may have multiple clients to give . . .
                // some protocols may have none and just use the socket as
                // a way of waking up the main thread (i.e. gftp)
			if ( !NestFd_isset( sock, &r_readyfds ) ) {
				continue;
			}

			nestClient *newClient = protocol_manager->accept( sock );
			NestFd_clear( sock, &r_readyfds );
			if ( NULL == newClient ) {
				continue;
			}

				// add new connection state to appropriate structures
			nest_server->NewClient( newClient );
        }

            // then queue all requests from existing clients
        nestClient *client;
        nestRequest *req;
        CLIENT_ITERATOR itr = nest_server->getClients()->begin();
        int size;
        if ( ( size = request_queue->clear() ) != 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
					   "Requests list was not previously cleared. Size %d.\n",
					   size );
        }
        while( itr != nest_server->getClients()->end( ) ) {
            client = *itr++;
            nest_debug( NEST_DEBUG_CLIENT, "Checking client %d (%s)\n",
						client->getId(), client->getUserIdentity() );

                // pass the r_readyfds to each client, if they are ready
                // they can create a request object
            req = req_handler->ReceiveRequest( client );
            if ( req != NULL ) {
                request_queue->addClient( client );
                nest_debug( NEST_DEBUG_CLIENT,
							"Client '%s' (%d) request: \"%s\"\n",
							client->getUserIdentity(),
							client->getId(),
							req->toString() );
            }
        } 

            // now iterate through new requests and handle them here
        req_handler->HandleRequests( request_queue, &readfds );

            // work on any io transfers last
            // this is important to be last to allow anticipatory scheduling
        io_manager->makeProgress( );

    } // end of main service loop

        // we should never get here
    nest_die( "Big trouble in Little China.\n" );
    return ERROR_EXIT;
}

static void
init( int argc, const char *argv[] )
{
    bool Daemonize = false;
	const char *value;

    setNewGroup();

        // set up our fd sets
    FD_ZERO( &writefds );
    FD_ZERO( &readfds  );

        // turn off all debugging unless instructed otherwise
    set_nest_debug_level( NEST_DEBUG_NEVER );

    set_nest_debug_level( NEST_DEBUG_MIN );
	nest_debug( NEST_DEBUG_ALWAYS,
				"NeST Server version '" NEST_VERSION __DATE__ "' starting\n" );

        // set up config file must occur before initing other stuff
    ReadConfigFile(argc, argv);

	// Instantiate and initialize the server object
	nest_server = new nestServer();
    nest_server->init( argc, argv );

        // read config values
	nest_debug( NEST_DEBUG_ALWAYS, "Checking debug setting\n" );
    if (GetConfigValue( Config_Debug, &value)) {
		nest_debug( NEST_DEBUG_ALWAYS, "%s\n", value );
        if ( add_nest_debug_level(value) != NEST_SUCCESS ) {
            nest_die( "Unknown debug value specified: %s\n", value );
        }
    }

    if ( GetConfigValue( Config_PrintHelp, &value)) {
        PrintHelp(value);
        exit( 0 );
    }

    if ( GetConfigValue( Config_DumpConfigFile, NULL)) {
        DumpSampleConfigFile();
        exit( 0 );
    }

    if ( GetConfigValue(Config_MaintenanceInterval, &value ) ) {
		nest_server->setMaintenanceInterval( atoi(value) );
    }

    if ( GetConfigValue( Config_ProcessUser, &value ) ) {
        if ( NestChangeUser( value ) < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Unable to switch to user %s\n", value );
        }
    }

    if ( GetConfigValue( Config_Daemonize, NULL ) ) {
        Daemonize = true;
        daemon_init( "NeST", LOG_USER );
        //debug_init( true );
    }
    
    if ( GetConfigValue( Config_DisableAuthentication, NULL ) ) {
		req_handler->setAllowedAuth( NEST_AUTH_NONE );
        nest_warn( __FILE__, __FUNCTION__, 
                "Running with authentication disabled is strongly "
                "discouraged!!\n" );
    }

    /*if ( !Daemonize )*/ debug_init( false );

        // install signal handlers
    installSignalHandlers( );

}

static void
debug_init( bool Daemonize )
{

    const char *log_file = NULL;

    if ( Daemonize ) {
        nest_debug_init( NULL, true );
     } else {
        nestProtocol *logfile = NULL;
        if( GetConfigValue( Config_LogFile, &log_file ) ) {
            assert( log_file != NULL );
            nest_printf( "Logging to %s.\n", log_file );
            logfile = storage_manager->openMetaFile( (char *)log_file, 
													 APPEND_FLAGS );
            assert( logfile != NULL );
        }
        if ( GetConfigValue( Config_DebugStderr, NULL ) ) {
            if ( logfile ) {
                nest_printf( "Logging to stderr and ignoring conflicting "
                        "directive to log to %s.\n", log_file );
            }
            logfile = NULL;
        }
        nest_debug_init( logfile, false );
    }
}

#define MAXFD 64

static void
daemon_init( const char *pname, int facility )
{
    pid_t pid = 0;

    if ((  pid = fork() ) < 0 ) {
        nest_die( "fork failed: %s", strerror( errno ) );
    } else if ( pid > 0 ) {
            // parent
        exit( 0 );
    }

        // first child continues
    setsid();

    NestSignal( SIGHUP, SIG_IGN ); // ignore sig_pipes

        // forking twice is not working
    if ( ( pid = fork() ) <  0 ) {
        nest_die( "fork failed: %s", strerror( errno ) );
    } else if ( pid > 0 ) {
            // first child terminates
        exit( 0 );
    }

    umask( 0 );     // clear our file mode creation mask . . . ?

        // should we attempt to close all open descriptors . . . ?
    fprintf( stderr, "NeST has been daemonized.\n" );
    for( int i = 3; i < MAXFD; i++ ) {
        close( i );
    }
        // should we open /dev/null for reading and writing and dup it 
        // to stdin and stdout
    int devnull = NestOpen( "/dev/null", O_RDWR ); 
    NestDup2( devnull, STDIN_FILENO );
    NestDup2( devnull, STDOUT_FILENO );
        
    openlog( pname, LOG_PID | LOG_CONS, facility );
}

/*
void
sigInt( int signal ) {
    nest_printf( "Exiting due to SIGINT.\n" );
    _exit( 0 );
}
*/

// ignore sigpipes
// pass sigchilds to the I/O manager
static void
installSignalHandlers( )
{

    NestSignal( SIGPIPE, SIG_IGN ); // ignore sig_pipes
 //   NestSignal( SIGINT,  sigInt );

    struct sigaction sa;
    sa.sa_sigaction = sigChild;
    sigemptyset( &sa.sa_mask );
    sa.sa_flags = SA_SIGINFO;

    if ( sigaction( SIGCHLD, &sa, 0 ) != 0 ) {
        nest_die( "sigaction: %s", strerror( errno ) );
    }
    if ( sigaction( SIGCLD, &sa, 0 ) != 0 ) {
        nest_die( "sigaction: %s", strerror( errno ) );
    }
}

static void
sigChild( int signal, siginfo_t *info, void *context )
{
    io_manager->sigChild( signal, info, context );    
}

static void
setNewGroup( void )
{
        // set ourselves up as a new group
    // doesn't work   arghh!
# if 0
    NestSetpgrp();
    if ( setpgid( 0, 37 ) ) {
        perror( "setpgid" );
    }
    if ( setsid() < 0 ) {
        perror( "setsid" );
    }
# endif
}

void
NestFdRegister( int fd, FD_TYPE type )
{
	if ( type & FD_READ ) {
		NestFd_set( fd, &readfds );
	}
	if ( type & FD_WRITE ) {
		NestFd_set( fd, &writefds );
	}
}

void
NestFdUnregister( int fd, FD_TYPE type )
{
	if ( type & FD_READ ) {
		NestFd_clear( fd, &readfds );
	}
	if ( type & FD_WRITE ) {
		NestFd_clear( fd, &writefds );
	}
}

bool
NestFdIsRegistered( int fd, FD_TYPE type )
{
	if ( type & FD_READ ) {
		return NestFd_isset( fd, &readfds );
	}
	if ( type & FD_WRITE ) {
		return NestFd_isset( fd, &writefds );
	}
	return 0;
}

void
NestFdSet( int fd, FD_TYPE type )
{
	if ( type & FD_READ ) {
		NestFd_set( fd, &r_readyfds );
	}
	if ( type & FD_WRITE ) {
		NestFd_set( fd, &w_readyfds );
	}
}

void
NestFdClear( int fd, FD_TYPE type )
{
	if ( type & FD_READ ) {
		NestFd_clear( fd, &r_readyfds );
	}
	if ( type & FD_WRITE ) {
		NestFd_clear( fd, &w_readyfds );
	}
}

bool
NestFdIsSet( int fd, FD_TYPE type )
{
	if ( type & FD_READ ) {
		return NestFd_isset( fd, &r_readyfds );
	}
	if ( type & FD_WRITE ) {
		return NestFd_isset( fd, &w_readyfds );
	}
	return 0;
}
