/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_proxy.h"

#include "nest_io_manager.h"
#include "nest_collection_manager.h"
#include "nest_storage_manager.h"
#include "nest_client_pool.h"
#include "nest_protocol_manager.h"
#include "nest_advertiser.h"
#include "nest_protocol_advertise.h"
#include "nest_request_handler.h"
#include "nest_config.h"
#include "nest_hash_table.h"
#include "nest_user.h"
#include "nest_main.h"
#include "nest_version.h"
#include "nest_server.h"

#include <sys/utsname.h>
#include <rpc/rpc.h>

    // time between any regular maintenance (in millisecs)
#define MAINTENANCE_INTERVAL 100000

    // the pid of the main server thread
	// used to make sure that only main server thread performs
	// protected duties
extern int admin_pid;

	// Storage manager, etc -- these are all global pointers
nestStorageManager		*storage_manager = NULL;
nestIdentityManager		*identity_manager = NULL;
nestFileCatalog			*file_catalog = NULL;
nestLotManager			*lot_manager = NULL;
nestCollectionManager	*collection_manager = NULL;
nestAdvertiser			*advertiser = NULL;
nestIoManager			*io_manager = NULL;
nestProtocolManager		*protocol_manager = NULL;
nestRequestHandler		*req_handler = NULL;

static const char* NestVersionString =
"$NestVersion: " NEST_VERSION " " __DATE__ " $";

nestServer::nestServer( void )
{
	maintenance_interval = MAINTENANCE_INTERVAL;
	clients = NULL;
}

nestServer::~nestServer( void )
{
}

int
nestServer::init( int argc, const char *argv[] )
{

        // instantiate our client pool and our serverTasks
        // these need to be done before args, bec if we're a proxy,
        // we're gonna add the proxy monitor as a client
    clients     = new nestClientPool( );

    initCommonMutex();

        // set admin pid value, must be done after daemonizing
    admin_pid = NestGetpid();
    nest_debug( NEST_DEBUG_ADMIN, "Admin pid is %d\n", admin_pid );

        // debug_init must be called after storage and before io
        //   unless Daemonize is true
        // protocolManager must proceed storage_manager
	storage_manager = new nestStorageManager( );
	identity_manager = new nestIdentityManager( );
	file_catalog = new nestFileCatalog( );
	lot_manager = new nestLotManager( );
    collection_manager = new nestCollectionManager( );
	req_handler = new nestRequestHandler( );

	// Ok, now initialize the storage manager, which will
	// initialize the others
	storage_manager->init( );

	// Create the protocol manager
	protocol_manager = new nestProtocolManager( );
    protocol_manager->init( );

	// And, the advertiser
	advertiser = new nestAdvertiser( );

	// The io manager
	io_manager = new nestIoManager( );
    io_manager->init( );

	// Initialize the proxy logic
    initProxy( clients ); 

        // add each available socket to our read set
        // the protocol manager has listened on one socket for each protocol
    for( int protocol_socket = protocol_manager->firstProtocolSocket();
         protocol_socket != protocol_manager->lastProtocolSocket();
         protocol_socket = protocol_manager->nextProtocolSocket() )
    {
        if ( protocol_socket > 0 ) {
			nest_debug( NEST_DEBUG_ALWAYS,
						"Adding socket %d (for %s) to read set\n",
						protocol_socket,
						protocol_manager->getProtocolNameFromSocket(protocol_socket ) );
			NestFdRegister( protocol_socket, FD_READ );
        }
    }

        // ask the ad manager to write the initial ad
    WriteAd();


    return 0;
}

// the max of the clients, the io and the protocols
// now that we add rpc fds to our select loop, I'm not sure how to set
// the maxfd . . . 
int
nestServer::getMaxfd( void )
{
    int maxfd = max( io_manager->getMaxFd(), clients->getMaxfd() );
    maxfd = max( maxfd, protocol_manager->maxSocket( ) );
    return maxfd;
}

NestReplyStatus
nestServer::NewClient( nestClient *newclient )
{
    int size = getClients()->addClient( newclient );
    nest_debug( NEST_DEBUG_CLIENT,
				"added a new client (has sock = %s),"
				" client count is %d\n",
				newclient->hasSocket() ? "true" : "false", size );

    if ( newclient->hasSocket() ) {
        nestProtocol *connection = newclient->getConnection();
            // may be an optimization to add to ready_fds
		nest_debug( NEST_DEBUG_FD, "setting read FD\n" );
		NestFdRegister( connection->getSocket(), FD_READ );
            // I don't think this is an optimization
        //NestFd_set( connection->getSocket(), &r_readyfds );
    }
    return NEST_SUCCESS;
}

nestClientPool *
nestServer::getClients( void )
{
	return clients;
}

NestReplyStatus
nestServer::WriteAd( void )
{
    NestReplyStatus status;

    nestProtocol *ad = storage_manager->openMetaFile
        ( AD_NAME, WRITE_FLAGS | O_TRUNC );

    if ( ad == NULL ) {
        nest_die( "Unable to open metafile %s\n", AD_NAME );
        status = NEST_NONPRIVELEDGED_SERVER;
    } else {
		// status = ad->Sendn( "[\n", 2 ); this is new classads
        status = advertiser->writeAd( ad );

            // now write some server stuff
        char buffer[NEST_MESSAGESIZE];
        snprintf( buffer, NEST_MESSAGESIZE, 
                "Current_clients   = %d\n"
                "Maintenance_interval = %.2f\n",
                ( ( clients != NULL ) ? clients->size() : 0 ),
                (float)maintenance_interval / THOUSAND );

        if ( serverIsProxy() ) {
            writeProxyAd( ad );
        }

		// this is new classads
		//snprintf( &buffer[NestStrlen(buffer)], NEST_MESSAGESIZE, "]\n" ); 
        snprintf( &buffer[NestStrlen(buffer)], NEST_MESSAGESIZE, "\n" ); 
        
        ad->Sendn( buffer, NestStrlen( buffer ) );
        delete( ad );
    }

    return status;
}

NestReplyStatus
nestServer::advertise( void )
{
    NestReplyStatus status = NEST_SUCCESS;


        // update ad
    WriteAd();
            
    const char *condor_advertise;
    char command[NEST_MESSAGESIZE];
    int retval;

        // condor_advertise
    if ( GetConfigValue( Config_CondorAdvertise, &condor_advertise ) ) {
        snprintf( command, NEST_MESSAGESIZE, "%s UPDATE_STORAGE_AD %s",
            condor_advertise, storage_manager->metaFilepath( AD_NAME ) );
        retval = system( command );
        nest_debug( NEST_DEBUG_SERVER, "Exec'd system %s: %d\n",
                command, retval );
        if ( retval < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Unable to exec command \"%s\"\n", command );
        }
    }

    return status;
}

NestReplyStatus 
nestServer::RegularMaintenance( void )
{
    NestReplyStatus status = NEST_SUCCESS;

    nest_debug( NEST_DEBUG_SERVER, "Performing regular maintenance now\n" );
    NestGettimeofday( &last_maintenance );

        // only do io::makeProgress for now
        // TODO need slow timeouts for regular maintenance
        // and fast timeouts for scheduling
        // turn off anticipatory scheduling for now

        // for a while with the proxies, this was needed bec the pot pipe
        // was somehow maybe not waking up the server
        // but I think that was a race condition otherwise solved.
    //io_manager->makeProgress();

    advertise();

    storage_manager->maintenance();

    proxyMaintenance();

    return status;
}

struct timeval *
nestServer::getNextTimeout( void )
{
	return NestNextTimeout( maintenance_interval, &last_maintenance );
}

void
nestServer::setMaintenanceInterval( int value )
{
	maintenance_interval = value;
}

const char *getRootDir( void )
{
	return storage_manager->getRootDir( );
}

const char *getNestVersionStr( void )
{
	return NestVersionString;
}

