/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_USER_H
#define NEST_USER_H

#include <string>
using std::string;

#include "nest_storage_common.h"

const user_id_t	USERID_INVALID = 0xffffffff;

// Dirty flags for the user object
#define USERDIRTY_NONE				0x0000
#define USERDIRTY_IDENTITY			0x0001
#define USERDIRTY_OSNAME			0x0002
#define USERDIRTY_UID				0x0004
#define USERDIRTY_IDENTITY_CLEAN	0x0008
#define USERDIRTY_SPECIAL			0x0010
#define USERDIRTY_ANY				0x001f

// Simple "tri-state"; should be somewhere else (nest_common.h ???)
class nestTriState
{
  public:
	nestTriState( void )
		{ isSet = false; };
	nestTriState( bool value )
		{ Set( value ); };
	nestTriState( const nestTriState *copy )
		{ this->value = copy->value; this->isSet = copy->isSet; };
	~nestTriState( void ) { };
	bool Set( bool value )
		{ this->value = value; isSet = true; return value; };
	bool Value( void ) const
		{ assert( isSet ); return value; }
	bool IsSet( void ) const
		{ return isSet; };
	void Reset( void )
		{ isSet = false; };
	
  private:
	bool	value;
	bool	isSet;
};

// Abstract base class for nestGroup and nestUser
class
nestUserOrGroup
{
  public:
	nestUserOrGroup( void );
	virtual ~nestUserOrGroup( void );
	virtual const char *getName( void ) const = 0;
	virtual const char *getType( void ) const = 0;
	virtual bool isEmpty( void ) const = 0; 
	virtual bool isGroup( void ) const = 0;
	virtual bool isUser( void ) const = 0;
	virtual bool match( const nestUserOrGroup *other ) const;
	virtual bool isAdmin( void ) const = 0;
	virtual user_id_t getId( void ) const = 0;

	// "Actual Name' string managment
	const char *getActual( void ) const;
	const char *setActual( const char *name ) const;
	const char *setActual( const string *name ) const;

  private:
	mutable char *actualName;

};

// Simple class to manage group info
class
nestGroup : public nestUserOrGroup
{
  public:
    nestGroup( const char *name);
    nestGroup( const string *name);
    nestGroup( const char *name, user_id_t gid );
    nestGroup( const string *name, user_id_t gid );
    nestGroup( const nestGroup *group);
	nestGroup( void );
    ~nestGroup( );

	// Set operations
	int setName( const char *name);
	int setName( const string *name);
	int setName( const char *name, user_id_t gid );
	int setName( const string *name, user_id_t gid );

	// Required from the base type
	bool isUser( void ) const { return false; };
	bool isGroup( void ) const { return true; };
	const char *getType( void ) const { return "Group"; };
	const char *getName( void ) const { return name; };
	bool isEmpty( void ) const { return ( name == NULL ); };
	user_id_t getGroupId( void ) const { return groupId; };
	user_id_t getId( void ) const { return groupId; };

	bool isNoGroup( void ) const;

	bool isAdmin( void ) const { return false; };

    bool match( const nestGroup *other ) const;
    bool match( const char *other ) const;
    bool match( const string *other ) const;
	bool match( const nestUserOrGroup *other ) const;

  private:
	char			*name;
	nestTriState	noGroup;
	user_id_t		groupId;

	int init( void );
};

// More complex class to manage NeST user info
class
nestUser : public nestUserOrGroup
{

  public:

	nestUser( const char *nestIdentity );
    nestUser( const string *nestIdentitiy );
	nestUser( const nestUser *);
	nestUser( void );
    ~nestUser( );

	int reset( void );
	int copyUser( const nestUser * );

	const char *getType( void ) const { return "User"; };
	bool isUser( void ) const { return true; };
	bool isGroup( void ) const { return false; };

	// Basic accessors
    const char *getName( void ) const
		{
			return getIdentity( );
		};
    const char *getOsName( void ) const
		{
			if ( isDirty(USERDIRTY_OSNAME) && !osName ) {
				fillOsName( );
			}
			return osName;
		};
    const char *getLocalName( void ) const;
    const char *getIdentity( void ) const
		{
			if ( isDirty(USERDIRTY_IDENTITY) && !identity ) {
				fillIdentity( );
			}
			return identity;
		};
    const char *getIdentityClean( void ) const
		{
			if ( isDirty(USERDIRTY_IDENTITY_CLEAN) && !identityClean ) {
				fillIdentityClean( );
			}
			return identityClean;
		};
	bool isEmpty( void ) const { return ( identity == NULL ); };
	bool hasGroup( void ) const { return ( group != NULL ); };

	// UID operations
    user_id_t setUid( user_id_t uid );
    user_id_t getUid( void ) const
		{
			if ( isDirty(USERDIRTY_UID) && (userId == USERID_INVALID) ) {
				fillUid( );
			}
			return userId;
		};
	user_id_t getId( void ) const { return getUid( ); }; 

	// "Local name" operations
	int resetLocalName( void );
    int setLocalName( const char *local_name );
    int setLocalName( const string *local_name );

	// O/S User name
	int resetOsName( void );
    int setOsName( const char *username );
    int setOsName( const string *username );

	int resetIdentity( const char *identity = NULL );
    int setIdentity( const char *ident );
    int setIdentity( const string *ident );

	// Try to match the name to an idenity or local user name
    int setName( const char *name );
    int setName( const string *name );

	// Group operations
    int setGroup( const nestGroup * );
    nestGroup *getGroup( void ) const
		{ return group; };
	const char *getGroupName( void ) const;

    bool match( const nestUser *other ) const;
    bool match( const char *other ) const;
    bool match( const string *other ) const;
	bool match( const nestUserOrGroup *other ) const;
    bool matchAny( const char *name ) const;
    bool matchAny( const string *name ) const;

	// Check the current user to see if it has "special" traits
	bool isAdmin( void ) const
		{
			if ( isDirty(USERDIRTY_SPECIAL) && !admin.IsSet() ) {
				fillSpecial( );
			}
			return admin.Value();
		};
	bool isAnonymous( void ) const
		{
			if ( isDirty(USERDIRTY_SPECIAL) && !anonymous.IsSet() ) {
				fillSpecial( );
			}
			return anonymous.Value();
		};
	bool isEverybody( void ) const
		{
			if ( isDirty(USERDIRTY_SPECIAL) && !everybody.IsSet() ) {
				fillSpecial( );
			}
			return everybody.Value();
		};
	bool isSystem( void ) const
		{
			if ( isDirty(USERDIRTY_SPECIAL) && !system.IsSet() ) {
				fillSpecial( );
			}
			return system.Value();
		};

	// Check a "generic" user identity for "special" traits
	bool isSpecial( const char * ) const;
	bool isAdmin( const char * ) const;
	bool isAnonymous( const char * ) const;
	bool isEverybody( const char * ) const;
	bool isSystem( const char * ) const;

  private:
    const char  *osName;		// UNIX user name
	const char  *localName;		// Name local to NeST
    const char  *identity;		// Identity string
    const char  *identityClean;
    user_id_t   userId;
	bool		userIdSet;
	nestGroup   *group;
	unsigned    dirty;

	// Admin / anonymous / everybody
	nestTriState	admin;		// Admin user?
	nestTriState	anonymous;	// Anonymous?
	nestTriState	everybody;	// Everybody?
	nestTriState	system;		// System user?

	// Private methods
	int init( void );
	void freeAll( void );

	// Fill in unknown pieces from known ones
	// Although these are declared as "const", they're really not
	int fillOsName( void ) const;
	int fillUid( void ) const;
	int fillIdentity( void ) const;
	int fillIdentityClean( void ) const;
	int fillSpecial( void ) const;
	int fillAll( void ) const;

	// Is the state used to generate these "dirty?"
	bool isDirty( unsigned which ) const
		{
			return ( dirty & which ) ? true : false;
		};
	void setDirty( unsigned which )
		{
			dirty |= which;
		};
	void unsetDirty( unsigned which )
		{
			dirty &= (~which);
		};
};

#endif
